chili = {
            'background_color': '#d63e3e',
            'eyeball_color': '#ffffff',
            'iris_color': '#4db818',
            'eye_size': 150,
            'eye_height': 80,
            'eye_separation': 450,
            'iris_size': 90,
            'pupil_scale': 0.8,
            'eye_shine': True,
            'eyelid_color': '#d63e3e',
            'nose_color': '#e38fae',
            'nose_vertical_position': 10,
            'nose_width': 80,
            'nose_height': -25,
            'mouth_color': '#222222',
            'mouth_width': 480,
            'mouth_height': 30,
            'mouth_thickness': 18,
            'brow_color': '#222222',
            'brow_width': 150,
            'brow_height': 120,
            'brow_thickness': 20
        }

gingerbreadman = {
            'background_color': '#8c6f5a',
            'eyeball_color': '#ffffff00',
            'iris_color': '#ffffff',
            'eye_size': 150,
            'eye_height': 80,
            'eye_separation': 300,
            'iris_size': 30,
            'pupil_scale': 0,
            'eye_shine': True,
            'eyelid_color': '#8c6f5a',
            'nose_color': '#9bbce0',
            'nose_vertical_position': 0,
            'nose_width': 0,
            'nose_height': 0,
            'mouth_color': '#9e0216',
            'mouth_width': 220,
            'mouth_height': 30,
            'mouth_thickness': 18,
            'brow_color': '#7bbac9',
            'brow_width': 70,
            'brow_height': 120,
            'brow_thickness': 20
        }

cutie = {
            'background_color': '#cccccc',
            'eyeball_color': '#ffffff',
            'iris_color': '#43d932',
            'eye_size': 150,
            'eye_height': 80,
            'eye_separation': 400,
            'iris_size': 90,
            'pupil_scale': 0.8,
            'eye_shine': True,
            'eyelid_color': '#cccccc',
            'nose_color': '#9bbce0',
            'nose_vertical_position': 0,
            'nose_width': 0,
            'nose_height': 0,
            'mouth_color': '#222222',
            'mouth_width': 320,
            'mouth_height': 30,
            'mouth_thickness': 18,
            'brow_color': '#222222',
            'brow_width': 110,
            'brow_height': 120,
            'brow_thickness': 20
        }


"""``speech`` contains the Python functions for interacting with the PyLips face.

The RobotFace class is the main class that is used to interact with the robot's face. 
It contains methods for speaking text, saving speech to a file, and stopping speech.

The PollyTTS and SystemTTS classes are used to generate speech audio files from text. 
The RobotFace class uses these classes to generate speech audio files and visemes. As
future systems are developed, additional TTS classes can be added to this module.

.. autosummary::
    :toctree:

    pylips.speech.RobotFace
    pylips.speech.PollyTTS
    pylips.speech.SystemTTS

"""

from pylips.speech.polly_tts import PollyTTS
from pylips.speech.system_tts import SystemTTS
from pylips.speech.robot_face import RobotFace

__all__ = [
    "RobotFace",
    "PollyTTS",
    "SystemTTS",
]


import boto3
import json
import os
from contextlib import closing
import pickle

class PollyTTS:
    '''
    A text-to-speech backend that uses Amazon Polly.

    This class is used to generate audio files from text using Amazon Polly. It can also generate visemes
    that correspond to the audio files that it generates.

    args:
        None
    '''

    def __init__(self):
        self.tts = boto3.client('polly')

        # all current amazon polly voices
        self.voices = ['Zeina','Hala','Zayd','Lisa','Arlet','Hiujin','Zhiyu','Naja','Mads',
                       'Sofie','Laura','Lotte','Ruben','Nicole','Olivia','Russell','Amy','Emma',
                       'Brian','Arthur','Aditi','Raveena','Kajal','Niamh','Aria','Ayanda','Danielle',
                       'Gregory','Ivy','Joanna','Kendra','Kimberly','Salli','Joey','Justin','Kevin',
                       'Matthew','Ruth','Stephen','Geraint','Suvi','Celine','Léa','Mathieu','Rémi',
                       'Isabelle','Chantal','Gabrielle','Liam','Marlene','Vicki','Hans','Daniel',
                       'Hannah','Aditi','Kajal','Dora','Karl','Carla','Bianca','Giorgio','Adriano',
                       'Mizuki','Takumi','Kazuha','Tomoko','Seoyeon','Liv','Ida','Ewa','Maja','Jacek',
                       'Jan','Ola','Camila','Vitoria','Ricardo','Thiago','Ines','Cristiano','Carmen',
                       'Tatyana','Maxim','Conchita','Lucia','Enrique','Sergio','Mia','Andrés','Lupe',
                       'Penelope','Miguel','Pedro','Astrid','Elin','Filiz','Burcu','Gwyneth']

    def list_voices(self):
        '''
        Lists all the voices that are available in the Amazon Polly TTS backend.

        For a more in-depth look at the voices, see the `Amazon Polly documentation 
        <https://docs.aws.amazon.com/polly/latest/dg/voicelist.html>`_.

        args:
            None
        '''
        for i, voice in enumerate(self.voices):
            print(f'{i}: {voice}')

    def gen_audio_and_visemes(self, text, voice_id=None, fname=None):
        '''
        Generates an audio file and visemes from a string of text using Amazon Polly.

        args:
            text (str): the text that the robot should speak
            voice_id (str): the voice that the robot should speak in
            fname (str): the name of the file that the audio should be saved to

        returns:
            (tuple): a tuple containing ``fname``, ``times``, and ``visemes``. fname is 
            the path to the audio file, times is a list of times that correspond to the initiation
            of the visemes, and visemes is a list of visemes that correspond to the words in the audio 
        '''
        if voice_id is None:
            voice_id = 'Justin'
            
        if fname is None:   
            fname = f"pylips_phrases/{voice_id}_output.mp3"
        else:
            #if it was already cached, return it, otherwise, generate it and return it
            fname = f"pylips_phrases/{fname}.mp3"
            if os.path.exists(fname):
                times, visemes = pickle.load(open(f'{fname[:-4]}.pkl', 'rb'))
                return fname, times, visemes

        # Synthesize speech
        response = self.tts.synthesize_speech(
            TextType='ssml',
            Text=f"<speak>{text}</speak>",
            OutputFormat='mp3',
            VoiceId=voice_id
        )

        audio_stream = response["AudioStream"]

        if not os.path.isdir('pylips_phrases'):
            os.mkdir('pylips_phrases')
             
        with open(fname, "wb") as file:
            file.write(audio_stream.read())

        # Synthesize visemes
        # Synthesize speech with viseme output
        response = self.tts.synthesize_speech(
            TextType='ssml',
            Text=f"<speak>{text}</speak>",
            OutputFormat='json',
            SpeechMarkTypes=['viseme'],
            VoiceId=voice_id
        )
        
         # Process the response
        if 'AudioStream' in response:
            # Do something with the audio stream if needed
            with closing(response["AudioStream"]) as stream:
                        data = stream.read().decode('utf-8')
                        xSheet = data.split('\n')
                        xSheet = [json.loads(line) for line in xSheet if line != '']

        times = [x['time'] / 1000. for x in xSheet]
        visemes = [VIS2IPA[x['value']] for x in xSheet]

        pickle.dump((times, visemes), open(f'{fname[:-4]}.pkl', 'wb'))

        return fname, times, visemes
    
    def get_audio_and_visemes(self, fname):
        '''
        Loads presaved audio and visemes from a file.

        args:
            fname (str): the name of the file that the audio and visemes were saved to
        
        returns:
            (tuple): a tuple containing ``fname``, ``times``, and ``visemes``. fname is 
            the path to the audio file, times is a list of times that correspond to the initiation
            of the visemes, and visemes is a list of visemes that correspond to the words in the audio 
 
        
        raises:
            Exception: if the file does not exist
        '''
        #if it was already cached, return it, otherwise, raise an error
        fname = f"pylips_phrases/{fname}.mp3"

        if os.path.exists(fname):
            times, visemes = pickle.load(open(f'{fname[:-4]}.pkl', 'rb'))
            return fname, times, visemes
        else:
            raise Exception(f'phrase {fname} does not exist')
        
        
VIS2IPA = {"p": "BILABIAL",
            "f": "LABIODENTAL",
            "T": "INTERDENTAL",
            "s": "DENTAL_ALVEOLAR",
            "t": "DENTAL_ALVEOLAR",
            "S": "POSTALVEOLAR",
            "r": "POSTALVEOLAR",
            "k": "VELAR_GLOTTAL",
            "i": "CLOSE_FRONT_VOWEL",
            "u": "CLOSE_BACK_VOWEL",
            "@": "MID_CENTRAL_VOWEL",
            "a": "OPEN_FRONT_VOWEL",
            "e": "OPEN_FRONT_VOWEL",
            "E": "OPEN_FRONT_VOWEL",
            "o": "OPEN_BACK_VOWEL",
            "O": "OPEN_BACK_VOWEL",
            "sil": "IDLE"}


"""``face`` contains the code for rendering the face.

Most of the code in this package is not exposed to the user; however,
the user can access the FacePresets file with some pre-defined
face configurations to set the appearance of the face.

.. autosummary::
    :toctree:

    pylips.face.FacePresets

"""

from pylips.face import FacePresets


from flask import Flask, render_template, request
from flask_socketio import SocketIO, emit
import sys
import logging

app = Flask(__name__)
socketio = SocketIO(app, cors_allowed_origins="*") #allow communication on the same IP (main use case for this package)

app.static_folder = 'static' 

@app.route('/face')
@app.route('/face/<face_name>')
def index(face_name='default'):
    return render_template('index.html', face_name=face_name)

@app.route('/editor')
def editor():
    return render_template('editor.html')

@socketio.on('face_control')
def handle_message(message):
    print(f"received to {message['name']}: {message['action_type']}")
    emit('face_control', message, broadcast=True)

if __name__ == '__main__':
    #TODO: add argparse for host and port
    socketio.run(app, host='0.0.0.0', port=8000)
    


"""A simple library for web-based conversational agents."""

__author__ = "Interaction Lab"
__email__ = "dennler@usc.edu"
__version__ = "0.0.10"

from pylips import face, speech

__all__ = [
    "speech",
    "face",
]


#!/usr/bin/env python
"""The setup script."""

from setuptools import find_packages, setup

with open("README.md", encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md", encoding="utf-8") as history_file:
    history = history_file.read()

install_requires = [
    #web server
    "flask>=2.0.1",
    "flask-socketio>=5.1.1",
    "python-socketio>=5.11.1",
    "websocket-client>=1.7.0",
    "requests>=2.31.0",
    #TTS
    "boto3>=1.18.67",
    "pygame>=2.1.1",
    "py3-tts>=3.5",
    "allosaurus>=1.0.2",
    "soundfile>=0.12.1",
]

extras_require = {
    "dev": [
        "pip>=20.3",
        "pylint",
        "yapf",
        "pre-commit",

        # Testing
        "pytest==7.0.1",
        "pytest-cov==3.0.0",
        "pytest-benchmark==3.4.1",
        "pytest-xdist==2.5.0",

        # Documentation
        "myst-nb==0.17.1",
        "Sphinx==4.5.0",
        "sphinx-autobuild==2021.3.14",
        "sphinx-autodoc-typehints==1.18.2",
        "sphinx-codeautolink==0.12.1",
        "sphinx-copybutton==0.3.1",
        "sphinx-jinja2-compat==0.2.0",
        "sphinx-material==0.0.32",
        "sphinx-prompt==1.5.0",
        "sphinx-tabs==3.3.1",
        "sphinx-toolbox==3.1.0",
        "sphinxcontrib-applehelp==1.0.4",
        "sphinxcontrib-devhelp==1.0.2",
        "sphinxcontrib-htmlhelp==2.0.1",
        "sphinxcontrib-jsmath==1.0.1",
        "sphinxcontrib-qthelp==1.0.3",
        "sphinxcontrib-serializinghtml==1.1.5",
        "sphinx-rtd-theme==1.3.0rc1",

        # Distribution
        "bump2version==1.0.1",
        "wheel==0.40.0",
        "twine==4.0.2",
        "check-wheel-contents==0.4.0",
    ],
}

setup(
    author="Interaction Lab",
    author_email="dennler@usc.edu",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: User Interfaces",
        "Topic :: Communications :: Chat",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
    ],
    description=
    "An interface for embodied conversational interaction.",
    install_requires=install_requires,
    extras_require=extras_require,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords=["lipsynch", "robot", "interaction", "conversational", "agent"],
    name="pylips",
    packages=find_packages(include=["pylips", "pylips.*"]),
    python_requires=">=3.8.0",
    test_suite="tests",
    url="https://github.com/interaction-lab/PyLips",
    version="0.0.11",
    zip_safe=False,
)


import socketio
import boto3
from contextlib import closing
import json
import pickle
import pygame
import time
import os

from .polly_tts import PollyTTS
from .system_tts import SystemTTS

class RobotFace:
    '''
    The controllable interface for speech, gaze, and expressions. 
    
    Args:
        robot_name (str): the identity of the robot that should be speaking
        server_ip (str): the location of the server running the flask application.
        tts_method (str): the method that the robot should use to generate speech.
        voice_id (str): the voice that the robot should use to generate speech.
    '''
    def __init__(self,
                 robot_name: str = 'default',
                 server_ip: str = 'http://localhost:8000',
                 tts_method: str = 'system',
                 voice_id: str = None):

        if os.path.exists('pylips_phrases') == False:
            os.mkdir('pylips_phrases')

        self.name = robot_name
        self.voice_id = voice_id

        try:
            self.io = socketio.Client()
            self.io.connect(server_ip)
        except socketio.exceptions.ConnectionError as e:
            print(f'Error connecting to server at {server_ip}. Try running:\npython3 -m pylips.face.start')

        pygame.mixer.init()
        self.channel = 0

        tts_methods = ['polly', 'system']
        if tts_method not in tts_methods:
            raise Exception(f'parameter tts_method must be one of {tts_methods}')
        
        # TODO: implement other TTS things, but for now we will use polly
        if tts_method == 'system':
            self.tts = SystemTTS()
        if tts_method == 'polly':
            self.tts = PollyTTS()

    def say(self, content, wait=False):
        '''
            Issues a speech command to the face.

            This method will take in a string of text and convert it to speech. This method
            may also shift ``self.channel`` to avoid competing with other connected faces.

            Args:
                content (str): the text that the robot should speak
                wait (bool): whether or not the function should wait for the speech to finish
        '''

        fname, times, visemes = self.tts.gen_audio_and_visemes(content, self.voice_id)

        request = { 
                'name': self.name,
                'action_type': 'say',
                'visemes': visemes,
                'times': times,
             }
        

        #load sound
        while pygame.mixer.Channel(self.channel).get_busy():
            self.channel += 1
             
        sound = pygame.mixer.Sound(fname)

        #play sound and face
        self.io.emit('face_control', request)
        pygame.mixer.Channel(self.channel).play(sound)

        while wait and pygame.mixer.Channel(self.channel).get_busy():
            pygame.time.wait(100)

    def save_file(self, content, tag):
        '''
            Saves a speech file to the disk.

            This method will take in a string of text and create the pkl and sound
            files necessary to call ``say_file`` later. 

            Args:
                content (str): the text that the robot should speak
                tag (str): the tag to refer to the file. This should be a string with
                    no extensions or folder paths -- e.g., 'statement1' or 'question2'
        '''
        self.tts.gen_audio_and_visemes(content, self.voice_id, tag)
         

    def say_file(self, tag):
        '''
            Plays a speech file that was saved to the disk.

            Args:
                tag (str): the tag that was used to save the file. This should be a string with
                    no extensions or folder paths -- e.g., 'statement1' or 'question2'
        '''

        fname, times, visemes = self.tts.get_audio_and_visemes(tag)

        request = { 
                'name': self.name,
                'action_type': 'say',
                'visemes': visemes,
                'times': times,
             }
        
        #load sound
        while pygame.mixer.Channel(self.channel).get_busy():
            self.channel += 1
             
        sound = pygame.mixer.Sound(fname)

        #play sound and face
        self.io.emit('face_control', request)
        pygame.mixer.Channel(self.channel).play(sound)


    def stop_speech(self):
        '''
        Cancels the current speech command.

        If the robot is speaking, the robot will stop speaking and the visemes will stop playing.
        If the robot is not speaking, this method will have no effect.

        Args:
            None
        '''
        request = { 
                'name': self.name,
                'action_type': 'stop_speech',
             }
        
        self.io.emit('face_control', request)
        pygame.mixer.Channel(self.channel).stop()
         
        
    def look(self, x, y, z, time):
        '''
        Looks to a location in the face's reference frame.

        The robots references frame has its origin at the center of the eyes, with the positive 
        x-axis pointing to the right, the positive y-axis pointing up, and the positive z-axis 
        pointing out of the face.

        TODO: add a check to make sure the location is within the robot's range of motion

        Args:
            x (float): the x-coordinate in millimeters
            y (float): the y-coordinate in millimeters
            z (float): the z-coordinate in millimeters
            time (int): the time in milliseconds that the robot should take to look to the location
        '''
        request = {
                'name': self.name,
                'action_type': 'look',
                'location': [x,y,z],
                'time': time
            }
        self.io.emit('face_control', request)

    def release_gaze(self):
        '''
        Cancels the current gaze command.

        If the robot is looking at a location, the robot will stop looking at that location and
        return to its default gaze behaviors, looking idly in different directions. If the robot
        is not looking at a location, this method will have no effect.

        Args:
            None
        '''
        request = {
            'name': self.name,
            'action_type': 'release_gaze',
        }
        self.io.emit('face_control', request)

    def express(self, aus, time):
        '''
        Activates a set of Action Units (AUs) on the face.

        The robot's face is controlled by a set of Action Units (AUs) that can be activated to
        create different facial expressions. This method will take in a dictionary of AUs that
        maps the AU number to the intensity that the AU should be activated at. The method will
        also take in a time in milliseconds to reach the requested AU state.

        Action units can also be unilateral, meaning that they can be activated on one side of the
        face. For example, AU1 is the inner brow raiser, and activating it on one side of the face
        can be done by specifying 'AU01l' or 'AU01r' in the dictionary.

        Args:
            aus (dict): a dictionary of AUs that maps the AU number to the intensity that the AU should 
                be activated at -- e.g., {'AU1': 1.0, 'AU2': 0.5, 'AU3': 0.0}
            time (int): the time in milliseconds that the robot should take to reach the requested AU state
        '''
        request = {
            'name': self.name,
            'action_type': 'express',
            'aus': aus,
            'time': time
        }

        self.io.emit('face_control', request)

    def set_appearance(self, config):
        '''
        Changes the design of the face.

        Different parameters can be passed to the face to change its appearance. This method will
        take in a dictionary of configuration parameters that will be used to change the face's
        appearance, and update the face to reflect the new configuration.

        Args:
            config (dict): a dictionary of configuration parameters that will be used to change the face's
                appearance -- e.g., {'iris_color': '#00FF00', 'eye_size': 200, 'eye_height': 60}
        '''
        request = {
            'name': self.name,
            'action_type': 'update_face',
            'configuration': config
        }
        self.io.emit('face_control', request)
        time.sleep(.3)

    def wait(self):
        '''
        Waits for the current speech command to finish.

        This method will wait for the current speech command to finish before returning. If the robot
        is not speaking, this method will return immediately.

        Args:
            None
        '''
        while pygame.mixer.Channel(self.channel).get_busy():
            pygame.time.wait(100)


         



import pyttsx3
from allosaurus.app import read_recognizer
import soundfile as sf
import pickle
import os

from sys import platform

class SystemTTS:
    '''
    A text-to-speech backend that uses the system's default TTS engine.

    args:
        None
    '''
    def __init__(self):
        self.engine = pyttsx3.init()
        self.engine.setProperty('rate', 120)
        self.model = read_recognizer()
        self.voices = [voice.id for voice in self.engine.getProperty('voices')]

    def list_voices(self):
        '''
        Lists all the voices that are available in the system's default TTS backend.
        
        args:
            None
        '''
        for i, voice in enumerate(self.voices):
            print(f'{i}: {voice}')

    def gen_audio_and_visemes(self, text, voice_id=None, fname=None):
        '''
        Generates audio and visemes from a string of text using the system's default TTS engine.

        args:
            text (str): the text that the robot should speak
            voice_id (str): the voice that the robot should speak in
            fname (str): the name of the file that the audio should be saved to
        
        returns:
            (tuple): a tuple containing ``fname``, ``times``, and ``visemes``. fname is 
            the path to the audio file, times is a list of times that correspond to the initiation
            of the visemes, and visemes is a list of visemes that correspond to the words in the audio 

        raises:
            Exception: if the voice_id is not in the list of available voices
        '''
        if voice_id is None:
            voice_id = 'en' if platform == "linux" or platform == "linux2" else 'default'
        
        elif type(voice_id) == int and voice_id < len(self.voices):
            voice_id = self.voices[voice_id]

        elif voice_id not in self.voices:
            raise Exception(f'voice "{voice_id}" does not exist')
        else:
            self.engine.setProperty('voice', voice_id)

        if fname is None:   
            fname = f"pylips_phrases/{voice_id}_output.wav"
        else:
            #if it was already cached, return it, otherwise, generate it and return it
            fname = f"pylips_phrases/{fname}.wav"
            if os.path.exists(fname):
                times, visemes = pickle.load(open(f'{fname[:-4]}.pkl', 'rb'))
                return fname, times, visemes

        # Synthesize speech
        if platform == "linux" or platform == "linux2":
            # linux has issues with saving multiple files...
            os.system(f"espeak-ng -v {voice_id} -s 100 '{text}' -w {fname}")
        else:
            self.engine.save_to_file(text, fname)
            self.engine.runAndWait()

        data, samplerate = sf.read(fname)
        sf.write(fname, data, samplerate)

        #synthesize visemes
        out = self.model.recognize(fname, timestamp=True, lang_id='eng')

        times = [i.split(' ')[0] for i in out.split('\n')]
        visemes = [IPA2VISEME[i.split(' ')[-1]] for i in out.split('\n')]


        times.append(len(data)/samplerate + 0.2)
        visemes.append('IDLE')

        pickle.dump((times, visemes), open(f'{fname[:-4]}.pkl', 'wb'))

        return fname, times, visemes
    
    def get_audio_and_visemes(self, fname):
        '''
        Gets the audio and visemes from a file that was previously generated.

        args:
            fname (str): the name of the file that the audio and visemes were saved to. It is not
                necessary to include the file extension.
        
        returns:
            (tuple): a tuple containing ``fname``, ``times``, and ``visemes``. fname is 
            the path to the audio file, times is a list of times that correspond to the initiation
            of the visemes, and visemes is a list of visemes that correspond to the words in the audio 
        '''
        #if it was already cached, return it, otherwise, raise an error
        fname = f"pylips_phrases/{fname}.wav"

        if os.path.exists(fname):
            times, visemes = pickle.load(open(f'{fname[:-4]}.pkl', 'rb'))
            return fname, times, visemes
        else:
            raise Exception(f'phrase {fname} does not exist')
        

IPA2VISEME = {
    'sil': 'IDLE',
    '': 'IDLE',
    
    'k͡p̚': 'BILABIAL', 
    'm': 'BILABIAL',
    'b': 'BILABIAL',
    'p': 'BILABIAL',
    'pʰ': 'BILABIAL',


    'v': 'LABIODENTAL',
    'f': 'LABIODENTAL',

    'θ': 'INTERDENTAL',
    'ð': 'INTERDENTAL',

    'l': 'DENTAL_ALVEOLAR',
    'd': 'DENTAL_ALVEOLAR',
    't': 'DENTAL_ALVEOLAR',
    'tʰ': 'DENTAL_ALVEOLAR',
    't̠': 'DENTAL_ALVEOLAR',
    'n': 'DENTAL_ALVEOLAR',
    'ɳ': 'DENTAL_ALVEOLAR',
    's': 'DENTAL_ALVEOLAR',
    'z': 'DENTAL_ALVEOLAR',
    
    'ʃ': 'POSTALVEOLAR',
    'ʒ': 'POSTALVEOLAR',
    'ɹ̩': 'POSTALVEOLAR',
    'ɹ': 'POSTALVEOLAR',
    'r': 'POSTALVEOLAR',
    'ɻ': 'POSTALVEOLAR',
    'ɾ': 'POSTALVEOLAR',
    'dʒ': 'POSTALVEOLAR',
    'tʃ': 'POSTALVEOLAR',
    't͡ʃʲ': 'POSTALVEOLAR',
    'ij': 'POSTALVEOLAR',
    'tɕʰ': 'POSTALVEOLAR',
    'x': 'POSTALVEOLAR',
    'd̠': 'POSTALVEOLAR',

    'h': 'VELAR_GLOTTAL',
    'k': 'VELAR_GLOTTAL',
    'kʰ': 'VELAR_GLOTTAL',
    'g': 'VELAR_GLOTTAL',
    'ɡ': 'VELAR_GLOTTAL',
    'ŋ': 'VELAR_GLOTTAL',
    'h': 'VELAR_GLOTTAL',
    'ʔ': 'VELAR_GLOTTAL',

    'ɪ': 'CLOSE_FRONT_VOWEL',
    'I': 'CLOSE_FRONT_VOWEL',
    'iː': 'CLOSE_FRONT_VOWEL',
    'ɪ̯': 'CLOSE_FRONT_VOWEL',
    'j': 'CLOSE_FRONT_VOWEL',
    'e': 'CLOSE_FRONT_VOWEL',
    'i': 'CLOSE_FRONT_VOWEL',
    'eː': 'CLOSE_FRONT_VOWEL',
    'e̞': 'CLOSE_FRONT_VOWEL',
    'øː': 'CLOSE_FRONT_VOWEL',

    'ɛ': 'OPEN_FRONT_VOWEL',
    'a': 'OPEN_FRONT_VOWEL',
    'æ': 'OPEN_FRONT_VOWEL',
    'ɛː': 'OPEN_FRONT_VOWEL',
    'aː': 'OPEN_FRONT_VOWEL',

    'ə': 'MID_CENTRAL_VOWEL',
    'ɚ': 'MID_CENTRAL_VOWEL',
    'ɐ': 'MID_CENTRAL_VOWEL',
    'ɐː': 'MID_CENTRAL_VOWEL',
    'ɘ': 'MID_CENTRAL_VOWEL',
    'əː': 'MID_CENTRAL_VOWEL',
    'ɜː': 'MID_CENTRAL_VOWEL',
    'ɵː': 'MID_CENTRAL_VOWEL',
    
    'w': "CLOSE_BACK_VOWEL",
    'ʊ': 'CLOSE_BACK_VOWEL',
    'u': 'CLOSE_BACK_VOWEL',
    'uː': 'CLOSE_BACK_VOWEL',
    'ʉ': 'CLOSE_BACK_VOWEL',
    'ʉː': 'CLOSE_BACK_VOWEL',
    'ɯ': 'CLOSE_BACK_VOWEL',
    'ʍ': 'CLOSE_BACK_VOWEL',

    'o': 'OPEN_BACK_VOWEL',
    'oː': 'OPEN_BACK_VOWEL',
    'ɔ': 'OPEN_BACK_VOWEL',
    'ɔː': 'OPEN_BACK_VOWEL',
    'ɑ': 'OPEN_BACK_VOWEL',
    'ɑː': 'OPEN_BACK_VOWEL',
    'ɒː': 'OPEN_BACK_VOWEL',
    'ɒ': 'OPEN_BACK_VOWEL',
    'ʌ': 'OPEN_BACK_VOWEL',   
}